using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Dynamic2D
{
    public struct Sprite
    {
        public Texture2D texture;
        public Rectangle sourceRect;
        public Vector2 vOrigin;
        public Color color;

        public Sprite(Sprite _copy)
        {
            texture = _copy.texture;
            sourceRect = _copy.sourceRect;
            vOrigin = _copy.vOrigin;
            color = _copy.color;
        }

        public Sprite(Texture2D _texture)
            : this(_texture, _texture.Bounds, Color.White)
        {
        }

        public Sprite(Texture2D _texture, Color _color)
            : this(_texture, _texture.Bounds, _color)
        {
        }

        public Sprite(Texture2D _texture, Vector2 _vOrigin)
            : this(_texture, _texture.Bounds, _vOrigin, Color.White)
        {
        }

        public Sprite(Texture2D _texture, Vector2 _vOrigin, Color _color)
            : this(_texture, _texture.Bounds, _vOrigin, _color)
        {
        }

        public Sprite(Texture2D _texture, Rectangle _sourceRect)
            : this(_texture, _sourceRect, Color.White)
        {
        }

        public Sprite(Texture2D _texture, Rectangle _sourceRect, Color _color)
            : this(_texture, _sourceRect, new Vector2((float)_sourceRect.Width / 2, (float)_sourceRect.Height / 2), Color.White)
        {
        }

        public Sprite(Texture2D _texture, Rectangle _sourceRect, Vector2 _vOrigin)
            : this(_texture, _sourceRect, _vOrigin, Color.White)
        {
        }

        public Sprite(Texture2D _texture, Rectangle _sourceRect, Vector2 _vOrigin, Color _color)
        {
            texture = _texture;
            sourceRect = _sourceRect;
            vOrigin = _vOrigin;
            color = _color;
        }
    }

    public class Animation
    {
        public Sprite[] Sprites;
        public float fps;
        public bool bRepeat;

        public Animation(Texture2D _texture, int tilesX, int tilesY, float _fps, bool _bRepeat = false)
        {
            int tileWidth = _texture.Width / tilesX;
            int tileHeight = _texture.Height / tilesY;
            Sprites = new Sprite[tilesX * tilesY];
            for (int y = 0; y < tilesY; y++)
            {
                for (int x = 0; x < tilesX; x++)
                {
                    Sprites[y * tilesX + x] = new Sprite(_texture, new Rectangle(x * tileWidth, y * tileHeight, tileWidth, tileHeight));
                }
            }

            fps = _fps;
            bRepeat = _bRepeat;
        }
}

    public struct AnimationInstance
    {
        public Animation Animation;
        public float PlaySpeed;
        public int Frame;
        public float Remainder;

        public AnimationInstance(Animation _Animation) :
            this(_Animation, 1.0f)
        {
        }

        public AnimationInstance(Animation _Animation, float _PlaySpeed)
        {
            Animation = _Animation;
            PlaySpeed = _PlaySpeed;
            Frame = 0;
            Remainder = 0.0f;
        }

        public void Update(TimeSpan timeSpan)
        {
            Remainder += (float)timeSpan.TotalSeconds * PlaySpeed;
            Frame = (Frame + (int)(Remainder * Animation.fps));
            if (Animation.bRepeat)
            {
                Frame %= Animation.Sprites.Length;
            }
            Remainder %= 1 / Animation.fps;
        }

        public void Draw(Game game, Vector2 position, float rotation, float scale, float layerDepth)
        {
            Debug.Assert(!HasEnded());
            game.DrawObject(Animation.Sprites[Frame], position, rotation, scale, layerDepth);
        }

        public bool HasEnded()
        {
            return Frame >= Animation.Sprites.Length;
        }
    }

    public class Box9
    {
        public Sprite boxSprite;
        public int x1;
        public int x2;
        public int x3;
        public int y1;
        public int y2;
        public int y3;

        public Box9(Texture2D _texture, int _x1, int _x2, int _y1, int _y2)
        {
            boxSprite = new Sprite(_texture, Vector2.Zero);
            x1 = _x1;
            x2 = _x2;
            x3 = _texture.Width;
            y1 = _y1;
            y2 = _y2;
            y3 = _texture.Height;
        }

        public void Draw(Game game, Vector2 position, Vector2 size, float layerDepth)
        {
            boxSprite.sourceRect = new Rectangle(0, 0, x1, y1);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X, (int)position.Y, x1, y1), layerDepth);
            boxSprite.sourceRect = new Rectangle(x1, 0, x2 - x1, y1);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X + x1, (int)position.Y, (int)size.X - x1 - (x3 - x2), y1), layerDepth);
            boxSprite.sourceRect = new Rectangle(x2, 0, x3 - x2, y1);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X + (int)size.X - (x3 - x2), (int)position.Y, (x3 - x2), y1), layerDepth);

            boxSprite.sourceRect = new Rectangle(0, y1, x1, y2 - y1);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X, (int)position.Y + y1, x1, (int)size.Y - y1 - (y3 - y2)), layerDepth);
            boxSprite.sourceRect = new Rectangle(x1, y1, x2 - x1, y2 - y1);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X + x1, (int)position.Y + y1, (int)size.X - x1 - (x3 - x2), (int)size.Y - y1 - (y3 - y2)), layerDepth);
            boxSprite.sourceRect = new Rectangle(x2, y1, x3 - x2, y2 - y1);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X + (int)size.X - (x3 - x2), (int)position.Y + y1, (x3 - x2), (int)size.Y - y1 - (y3 - y2)), layerDepth);

            boxSprite.sourceRect = new Rectangle(0, y2, x1, y3 - y2);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X, (int)position.Y + (int)size.Y - (y3 - y2), x1, (y3 - y2)), layerDepth);
            boxSprite.sourceRect = new Rectangle(x1, y2, x2 - x1, y3 - y2);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X + x1, (int)position.Y + (int)size.Y - (y3 - y2), (int)size.X - x1 - (x3 - x2), (y3 - y2)), layerDepth);
            boxSprite.sourceRect = new Rectangle(x2, y2, x3 - x2, y3 - y2);
            game.DrawObjectScreenSpace(boxSprite, new Rectangle((int)position.X + (int)size.X - (x3 - x2), (int)position.Y + (int)size.Y - (y3 - y2), (x3 - x2), (y3 - y2)), layerDepth);
        }
    }

    public class Game : Microsoft.Xna.Framework.Game
    {
        public static Vector2 Vector2FromAngle(float fAngle)
        {
            return new Vector2(
                (float)Math.Sin(fAngle),
                (float)-Math.Cos(fAngle));
        }

        public static float AngleFromVector2(Vector2 vector)
        {
            return (float)Math.Atan2(vector.X, -vector.Y);
        }

        GraphicsDeviceManager graphicsManager;
        public SpriteBatch spriteBatch { get; protected set; }

        public bool bPause = false;
        public TimeSpan GameTime = TimeSpan.Zero;

        public enum LayerId
        {
            Background,
            World,
            Foreground,
        }

        public struct Layer
        {
            public LinkedList<GameObject> GameObjects;
            public SamplerState SamplerState;
        }

        public Layer[] Layers = new Layer[3];

        public Random random = new Random();
        public GamePadState gamePadState;
        public GamePadState prevGamePadState;
        public KeyboardState keyboardState;
        public KeyboardState prevKeyboardState;
#if !XBOX
        public MouseState mouseState;
        public MouseState prevMouseState;
#endif

        public Vector2 cameraPosition = Vector2.Zero;

        public Game()
        {
            graphicsManager = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            Layer Background;
            Background.GameObjects = new LinkedList<GameObject>();
            Background.SamplerState = SamplerState.LinearWrap;
            Layers[0] = Background;

            Layer World;
            World.GameObjects = new LinkedList<GameObject>();
            World.SamplerState = SamplerState.LinearClamp;
            Layers[1] = World;

            Layer Foreground;
            Foreground.GameObjects = new LinkedList<GameObject>();
            Foreground.SamplerState = SamplerState.LinearClamp;
            Layers[2] = Foreground;
        }

        /// <summary>
        /// Allows the game to perform any initialization it needs to before starting to run.
        /// This is where it can query for any required services and load any non-graphic
        /// related content.  Calling base.Initialize will enumerate through any components
        /// and initialize them as well.
        /// </summary>
        protected override void Initialize()
        {
            graphicsManager.PreferredBackBufferWidth = 1280;
            graphicsManager.PreferredBackBufferHeight = 720;
            graphicsManager.PreferMultiSampling = true;
            graphicsManager.ApplyChanges();
#if XBOX
            GamerServicesDispatcher.WindowHandle = Window.Handle;
            GamerServicesDispatcher.Initialize(Services);
#endif
            base.Initialize();

            spriteBatch = new SpriteBatch(GraphicsDevice);

            SoundEffect.MasterVolume = 0.3f;
        }

        /// <summary>
        /// LoadContent will be called once per game and is the place to load
        /// all of your content.
        /// </summary>
        protected override void LoadContent()
        {
        }

        /// <summary>
        /// UnloadContent will be called once per game and is the place to unload
        /// all content.
        /// </summary>
        protected override void UnloadContent()
        {
        }

        /// <summary>
        /// Allows the game to run logic such as updating the world,
        /// checking for collisions, gathering input, and playing audio.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Update(GameTime gameTime)
        {
            base.Update(gameTime);
#if XBOX
            GamerServicesDispatcher.Update();
#endif
            prevGamePadState = gamePadState;
            gamePadState = GamePad.GetState(PlayerIndex.One, GamePadDeadZone.Circular);

            prevKeyboardState = keyboardState;
            keyboardState = Keyboard.GetState(PlayerIndex.One);
#if !XBOX
            prevMouseState = mouseState;
            mouseState = Mouse.GetState();
#endif

            if (IsActive
#if XBOX
                && !Guide.IsVisible && gamePadState.IsConnected
#endif
)
            {
                PreUpdateGame(gameTime);

                if (!bPause)
                {
                    GameTime += gameTime.ElapsedGameTime;

                    UpdateCollision(gameTime);

                    // Update all game objects
                    foreach (Layer layer in Layers)
                    {
                        for (LinkedListNode<GameObject> gameObjectNode = layer.GameObjects.First; gameObjectNode != null; gameObjectNode = gameObjectNode.Next)
                        {
                            // Game object could have been destroyed by a game object updated earlier
                            // don't update it if it has been
                            if (gameObjectNode.Value.IsActive())
                            {
                                gameObjectNode.Value.Update(gameTime);
                            }
                        }

                        // Remove any game objects that have been destroyed from the GameObjects list
                        LinkedListNode<GameObject> nextGameObjectNode;
                        for (LinkedListNode<GameObject> gameObjectNode = layer.GameObjects.First; gameObjectNode != null; gameObjectNode = nextGameObjectNode)
                        {
                            nextGameObjectNode = gameObjectNode.Next;
                            if (!gameObjectNode.Value.IsActive())
                            {
                                layer.GameObjects.Remove(gameObjectNode);
                                gameObjectNode.Value.PostRemoveFromGame();
                            }
                        }
                    }

                    UpdateGame(gameTime);
                }
            }
        }

        protected virtual void PreUpdateGame(GameTime gameTime)
        {
        }

        protected virtual void UpdateGame(GameTime gameTime)
        {
        }

        protected void UpdateCollision(GameTime gameTime)
        {
            foreach (Layer layer in Layers)
            {
                for (LinkedListNode<GameObject> gameObjectNode1 = layer.GameObjects.First; gameObjectNode1 != null; gameObjectNode1 = gameObjectNode1.Next)
                {
                    GameObject gameObject1 = gameObjectNode1.Value;
                    if (gameObject1.IsActive() && gameObject1.CanCollide())
                    {
                        //Rectangle gameObject1Rect = new Rectangle((int)(gameObject1.position.X - gameObject1.hotspot.LocalSpace().X), (int)(gameObject1.position.Y - gameObject1.hotspot.LocalSpace().Y), gameObject1.sprite.Width, gameObject1.sprite.Height);
                        for (LinkedListNode<GameObject> gameObjectNode2 = gameObjectNode1.Next; gameObjectNode2 != null; gameObjectNode2 = gameObjectNode2.Next)
                        {
                            GameObject gameObject2 = gameObjectNode2.Value;
                            if (gameObject2.IsActive() && gameObject2.CanCollide()
                                && gameObject1.CanCollideWith(gameObject2) && gameObject2.CanCollideWith(gameObject1))
                            {
                                //Rectangle gameObject2Rect = new Rectangle((int)(gameObject2.position.X - gameObject2.hotspot.LocalSpace().X), (int)(gameObject2.position.Y - gameObject2.hotspot.LocalSpace().Y), gameObject2.sprite.Width, gameObject2.sprite.Height);
                                //if (gameObject1Rect.Intersects(gameObject2Rect))
                                float totalRadius = (gameObject1.collisionRadius + gameObject2.collisionRadius);
                                if (Vector2.DistanceSquared(gameObject1.position, gameObject2.position) <= totalRadius * totalRadius)
                                {
                                    gameObject1.NotifyCollision(gameObject2);
                                    // only notify object 2 about the collision if both survive object 1 being notified
                                    if (gameObject1.IsActive() && gameObject2.IsActive())
                                    {
                                        gameObject2.NotifyCollision(gameObject1);
                                    }
                                    // if the first object has been destroyed in the collision then 
                                    if (!gameObject1.IsActive())
                                    {
                                        break;
                                    }
                                }
                            }
                        }
                    }
                }
            }
        }

        /// <summary>
        /// This is called when the game should draw itself.
        /// </summary>
        /// <param name="gameTime">Provides a snapshot of timing values.</param>
        protected override void Draw(GameTime gameTime)
        {
            graphicsManager.GraphicsDevice.Clear(Color.CornflowerBlue);

            foreach (Layer layer in Layers)
            {
                spriteBatch.Begin(SpriteSortMode.BackToFront, BlendState.AlphaBlend, layer.SamplerState, null, null);
                foreach (GameObject gameObject in layer.GameObjects)
                {
                    gameObject.Draw(gameTime);
                }
                spriteBatch.End();
            }

            base.Draw(gameTime);
        }

        //public void AddObjectToGame(GameObject gameObject, LayerId layer)
        //{
        //    Layers[layer].GameObjects.AddLast(gameObject.
        //}

        public void DrawObject(Sprite sprite, Vector2 position, float rotation, float scale, float layerDepth)
        {
            position = position - cameraPosition + new Vector2(GraphicsDevice.Viewport.Width / 2, GraphicsDevice.Viewport.Height / 2);
            spriteBatch.Draw(sprite.texture, position, sprite.sourceRect, sprite.color, rotation, sprite.vOrigin, scale, SpriteEffects.None, layerDepth);
        }

        public void DrawObject(Sprite sprite, Vector2 position, float rotation, Vector2 scale, float layerDepth)
        {
            position = position - cameraPosition + new Vector2(GraphicsDevice.Viewport.Width / 2, GraphicsDevice.Viewport.Height / 2);
            spriteBatch.Draw(sprite.texture, position, sprite.sourceRect, sprite.color, rotation, sprite.vOrigin, scale, SpriteEffects.None, layerDepth);
        }

        public void DrawObjectScreenSpace(Sprite sprite, Vector2 position, float rotation, float scale, float layerDepth)
        {
            spriteBatch.Draw(sprite.texture, position, sprite.sourceRect, sprite.color, rotation, sprite.vOrigin, scale, SpriteEffects.None, layerDepth);
        }

        public void DrawObjectScreenSpace(Sprite sprite, Vector2 position, float rotation, Vector2 scale, float layerDepth)
        {
            spriteBatch.Draw(sprite.texture, position, sprite.sourceRect, sprite.color, rotation, sprite.vOrigin, scale, SpriteEffects.None, layerDepth);
        }

        public void DrawObjectScreenSpace(Sprite sprite, Rectangle destRect, float layerDepth)
        {
            spriteBatch.Draw(sprite.texture, destRect, sprite.sourceRect, sprite.color, 0, sprite.vOrigin, SpriteEffects.None, layerDepth);
        }

        public void DrawBackground(Sprite sprite, float rotation, float layerDepth)
        {
            Rectangle DestRect = new Rectangle(0, 0, GraphicsDevice.Viewport.Width, GraphicsDevice.Viewport.Height);
            spriteBatch.Draw(sprite.texture, DestRect, sprite.sourceRect, sprite.color, rotation, sprite.vOrigin, SpriteEffects.None, layerDepth);
        }
    }
}
