﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;

namespace Dynamic2D
{
    public class GameObjectBase
    {
        public Type GameObjectType;

        public Sprite sprite;
        public Animation animation;

        public float layerDepth = 0.5f;
        public float collisionRadius;

        public T ConstructInstance<T>() where T : GameObject
        {
            Debug.Assert(GameObjectType == typeof(T) || GameObjectType.IsSubclassOf(typeof(T)));
            return (T)GameObjectType.GetConstructor(new Type[] { GetType() }).Invoke(new Object[] { this });
        }
    }

    public class GameObject
    {
        public bool bActive = false;
        protected Game game = null;
        protected LinkedListNode<GameObject> gameObjectNode = null;

        public GameObjectBase BaseObject;

        public Sprite sprite;
        public AnimationInstance animation;

        public float layerDepth = 0.5f;
        public float scale = 1.0f;
        public float collisionRadius = 1.0f;

        public Vector2 position = Vector2.Zero;
        public float rotation = 0.0f;
        public Vector2 velocity = Vector2.Zero;
        public float rotVelocity = 0.0f;

        public Vector2 TransformLocalToWorld(Vector2 local)
        {
            Vector2 rotated = new Vector2(
                (float)Math.Cos(rotation) * local.X +
                (float)-Math.Sin(rotation) * local.Y,
                (float)Math.Sin(rotation) * local.X +
                (float)Math.Cos(rotation) * local.Y);
            return rotated * scale + position;
        }

        public GameObject(GameObjectBase _BaseObject)
        {
            BaseObject = _BaseObject;
            gameObjectNode = new LinkedListNode<GameObject>(this);
        }

        // Called to add this newly-created or recycled object to the game
        public void AddToGame(Game _game, Game.LayerId layerId)
        {
            Debug.Assert(!bActive);
            Debug.Assert(game == null);
            bActive = true;
            game = _game;
            game.Layers[(int)layerId].GameObjects.AddLast(gameObjectNode);
            PostAddToGame();
        }

        // Called automatically after this object has been added to the game
        public virtual void PostAddToGame()
        {
            // Set up your game object here
            sprite = BaseObject.sprite;
            animation = new AnimationInstance(BaseObject.animation);

            layerDepth = BaseObject.layerDepth;
            collisionRadius = BaseObject.collisionRadius;
        }

        // Called to remove this object from the game
        public void RemoveFromGame()
        {
            bActive = false;

            // Don't remove from the GameObjects list here, as we are probably in the middle of iterating it.
            // The game will do it when it notices bActive = false
            //game.GameObjects.Remove(gameObjectNode);
        }

        // Called automatically after this object has been removed from the game
        public virtual void PostRemoveFromGame()
        {
            // Don't null the game pointer until after we have been removed from the game
            // that way we can use it to add ourself to a dead object cache
            // don't do that in RemoveFromGame() above 
            game = null;
        }

        public bool IsActive()
        {
            return bActive;
        }

        public virtual void Update(GameTime gameTime)
        {
            position += velocity * (float)gameTime.ElapsedGameTime.TotalSeconds;
            rotation += rotVelocity * (float)gameTime.ElapsedGameTime.TotalSeconds;

            if (animation.Animation != null)
            {
                animation.Update(gameTime.ElapsedGameTime);
            }
        }

        public virtual void Draw(GameTime gameTime)
        {
            if (animation.Animation != null)
            {
                animation.Draw(game, position, rotation, scale, layerDepth);
            }
            if (sprite.texture != null)
            {
                game.DrawObject(sprite, position, rotation, scale, layerDepth);
            }
        }

        public virtual bool CanCollide()
        {
            return collisionRadius > 0;
        }

        public virtual bool CanCollideWith(GameObject other)
        {
            return true;
        }

        public virtual void NotifyCollision(GameObject other)
        {
        }
    }
}
